#ifndef CREATURELIB_EXECUTINGATTACK_HPP
#define CREATURELIB_EXECUTINGATTACK_HPP

#include "Creature.hpp"

namespace CreatureLib::Battling {
    class ExecutingAttack : public ScriptSource {
    public:
        class HitData {
            bool _critical = false;
            u8 _basePower = 0;
            float _effectiveness = 1;
            u32 _damage = 0;
            u8 _type = 0;
            bool _hasFailed = false;

        public:
            HitData() noexcept {}

            [[nodiscard]] inline bool IsCritical() const noexcept { return _critical; }
            [[nodiscard]] inline u8 GetBasePower() const noexcept { return _basePower; }
            [[nodiscard]] inline float GetEffectiveness() const noexcept { return _effectiveness; }
            [[nodiscard]] inline u32 GetDamage() const noexcept { return _damage; }
            [[nodiscard]] inline u8 GetType() const noexcept { return _type; }
            [[nodiscard]] inline bool HasFailed() const noexcept { return _hasFailed; }

            inline void SetCritical(bool value) noexcept { _critical = value; }
            inline void SetBasePower(u8 value) noexcept { _basePower = value; }
            inline void SetEffectiveness(float value) noexcept { _effectiveness = value; }
            inline void SetDamage(u32 value) noexcept { _damage = value; }
            inline void SetType(u8 value) noexcept { _type = value; }
            inline void Fail() noexcept { _hasFailed = true; }
        };

    private:
        u8 _numberHits;
        std::unique_ptr<HitData[]> _hits;
        ArbUt::BorrowedPtr<Creature> _user;
        ArbUt::BorrowedPtr<LearnedAttack> _attack;
        ArbUt::BorrowedPtr<const Library::AttackData> _useAttack;
        ArbUt::OptionalUniquePtr<BattleScript> _script = nullptr;
        ArbUt::List<ArbUt::OptionalBorrowedPtr<Creature>> _targets;

    public:
        ExecutingAttack(const ArbUt::List<ArbUt::OptionalBorrowedPtr<Creature>>& targets, u8 numberHits,
                        ArbUt::BorrowedPtr<Creature> user, const ArbUt::BorrowedPtr<LearnedAttack>& attack,
                        const ArbUt::BorrowedPtr<const Library::AttackData>& useAttack, BattleScript* nullable script)
            : _numberHits(numberHits), _hits(std::make_unique<HitData[]>(targets.Count() * numberHits)), _user(user),
              _attack(attack), _useAttack(useAttack), _targets(targets) {
            // Take ownership of the script of the attack choice, and give attack choice our initial nullptr.
            _script = script;
        }
        ExecutingAttack(const ExecutingAttack&) = delete;
        ExecutingAttack& operator=(const ExecutingAttack&) = delete;

        virtual ~ExecutingAttack() noexcept = default;

        HitData& GetHitData(ArbUt::BorrowedPtr<Creature> creature, u8 hit) {
            for (u8 i = 0; i < _targets.Count(); i++) {
                if (!_targets[i].HasValue()) {
                    continue;
                }
                if (_targets[i].GetValue() == creature) {
                    return _hits[i * _numberHits + hit];
                }
            }
            THROW("Invalid target requested.");
        }

        HitData* non_null GetTargetIteratorBegin(ArbUt::BorrowedPtr<Creature> creature) {
            for (u8 i = 0; i < _targets.Count(); i++) {
                if (!_targets[i].HasValue()) {
                    continue;
                }
                if (_targets[i].GetValue() == creature) {
                    return &_hits[i * _numberHits];
                }
            }
            THROW("Invalid target requested.");
        }

        bool IsCreatureTarget(ArbUt::BorrowedPtr<Creature> creature) noexcept {
            for (u8 i = 0; i < _targets.Count(); i++) {
                if (!_targets[i].HasValue()) {
                    continue;
                }
                if (_targets[i].GetValue() == creature) {
                    return true;
                }
            }
            return false;
        }

        inline u8 GetTargetCount() const noexcept { return _targets.Count(); }

        inline const ArbUt::List<ArbUt::OptionalBorrowedPtr<Creature>>& GetTargets() const noexcept { return _targets; }
        inline u8 GetNumberOfHits() const noexcept { return _numberHits; }

        inline const ArbUt::BorrowedPtr<Creature>& GetUser() noexcept { return _user; }

        inline const ArbUt::BorrowedPtr<LearnedAttack>& GetAttack() noexcept { return _attack; }
        inline const ArbUt::BorrowedPtr<const Library::AttackData>& GetUseAttack() noexcept { return _useAttack; }
        size_t ScriptCount() const override { return _user->ScriptCount() + 1; }

        inline ArbUt::BorrowedPtr<BattleScript> GetScript() const noexcept { return _script.GetValue(); }

        void GetActiveScripts(ArbUt::List<ScriptWrapper>& scripts) override {
            GetOwnScripts(scripts);
            _user->GetActiveScripts(scripts);
        }
        void GetOwnScripts(ArbUt::List<ScriptWrapper>& scripts) override {
            scripts.Append(ScriptWrapper::FromScript(&_script));
        }
    };
}

#endif // CREATURELIB_EXECUTINGATTACK_HPP
