#ifndef CREATURELIB_BASELIBRARY_HPP
#define CREATURELIB_BASELIBRARY_HPP

#include <Arbutils/Collections/List.hpp>
#include <Arbutils/Collections/StringViewDictionary.hpp>
#include <Arbutils/Memory/Memory.hpp>
#include <Arbutils/Random.hpp>
#include <Arbutils/String/StringView.hpp>

namespace CreatureLib::Library {
    template <class T> class BaseLibrary {
    protected:
        ArbUt::StringViewDictionary<std::unique_ptr<const T>> _values;
        ArbUt::List<u32> _listValues;

    public:
        BaseLibrary(size_t initialCapacity = 32) : _values(initialCapacity), _listValues(initialCapacity) {}

        virtual ~BaseLibrary() noexcept { _values.Clear(); }

        inline virtual void Insert(const ArbUt::StringView& key, const T* non_null value) {
            EnsureNotNull(value)
            _values.GetStdMap().insert({key, std::unique_ptr<const T>(value)});
            _listValues.Append(key);
        }

        inline void Delete(const ArbUt::StringView& key) noexcept {
            _values.erase(key.GetHash());
            auto k = _listValues.IndexOf(key);
            if (k.has_value()) {
                _listValues.Remove(k.value());
            }
        }
        inline void DeleteByHash(u32 hashedKey) noexcept {
            _values.Remove(hashedKey);
            auto k = _listValues.IndexOf(hashedKey);
            if (k.has_value()) {
                _listValues.Remove(k.value());
            }
        }

        std::optional<ArbUt::BorrowedPtr<const T>> TryGet(const ArbUt::StringView& name) const noexcept {
            auto opt = _values.TryGet(name);
            if (opt.has_value()) {
                return std::make_optional<ArbUt::BorrowedPtr<const T>>(opt.value());
            }
            return {};
        }
        std::optional<ArbUt::BorrowedPtr<const T>> TryGetByHash(u32 hashedKey) const noexcept {
            auto opt = _values.TryGet(hashedKey);
            if (opt.has_value()) {
                return std::make_optional<ArbUt::BorrowedPtr<const T>>(opt.value());
            }
            return {};
        }

        [[nodiscard]] inline ArbUt::BorrowedPtr<const T> Get(const ArbUt::StringView& name) const {
            return _values.Get(name);
        }
        [[nodiscard]] inline ArbUt::BorrowedPtr<const T> GetByHash(u32 hashedKey) const {
            return _values.GetFromHash(hashedKey);
        }

        [[nodiscard]] inline ArbUt::BorrowedPtr<const T> operator[](const ArbUt::BasicStringView& name) const {
            return Get(name);
        }
        [[nodiscard]] inline ArbUt::BorrowedPtr<const T> operator[](u32 hashedKey) const { return Get(hashedKey); }
        [[nodiscard]] inline const ArbUt::StringViewDictionary<const std::unique_ptr<const T>>&
        GetIterator() const noexcept {
            return _values;
        }

        using const_iterator = typename ArbUt::StringViewDictionary<std::unique_ptr<const T>>::const_iterator;

        inline const_iterator begin() const { return _values.begin(); }

        inline const_iterator end() const { return _values.end(); }

        [[nodiscard]] size_t GetCount() const noexcept { return _values.Count(); }
        ArbUt::BorrowedPtr<const T> GetAtIndex(size_t index) const {
            return std::next(std::begin(_values), index)->second;
        }

        inline ArbUt::BorrowedPtr<const T> GetRandomValue(ArbUt::Random& rand) const noexcept {
            auto i = rand.Get(_listValues.Count());
            return _values[_listValues[i]];
        }
    };
}

#endif // CREATURELIB_BASELIBRARY_HPP
