#ifndef CREATURELIB_CREATURESPECIES_HPP
#define CREATURELIB_CREATURESPECIES_HPP

#include "../Gender.hpp"
#include "SpeciesVariant.hpp"

namespace CreatureLib::Library {
    /// @brief This holds the data required for a species of a creature, so the general data we can describe different
    //    creatures with.
    class CreatureSpecies {
        struct impl;
        std::unique_ptr<impl> _impl;

    public:
        /// @brief Instantiates a CreatureSpecies.
        /// @param id The unique id of the creature species.
        /// @param name The name of the creature species. This should generally not be its display name.
        /// @param defaultVariant The variant the creature species has by default.
        /// @param genderRatio The ratio at which a creature species is male or female. 0 is always female, 1 is always
        /// male. -1 makes the creature genderless.
        /// @param growthRate The name of the growthrate this creature species should follow. These should be defined in
        /// the growthrate library.
        /// @param captureRate The chance to capture the creature species, between 0 and 255. 255 means instant capture,
        /// 0 means impossible to capture.
        /// @param flags A set of flags for use by the developer. These can be used for easy grouping.
        CreatureSpecies(u16 id, const ArbUt::StringView& name, const SpeciesVariant* non_null defaultVariant,
                        float genderRatio, const ArbUt::StringView& growthRate, u8 captureRate,
                        std::unordered_set<u32> flags = {});

        virtual ~CreatureSpecies();

        /// @brief Returns the unique id of the creature species.
        /// @return The unique id of the creature species
        u16 GetId() const noexcept;
        /// @brief Returns the name of the species.
        /// @return The name of the species.
        [[nodiscard]] const ArbUt::StringView& GetName() const noexcept;
        /// @brief Returns the gender rate of the creature species.
        /// @return The gender rate of the creature species. 0 is always female, 1 is always male. -1 makes the creature
        /// genderless.
        float GetGenderRate() const noexcept;
        /// @brief Returns the growthrate name of the creature species, as defined in the growthrate library.
        /// @return The growthrate name of the creature species, as defined in the growthrate library.
        const ArbUt::StringView& GetGrowthRate() const noexcept;
        /// @brief Returns the capture rate of the creature species.
        /// @return The base capture rate of the creature species. 255 means instant capture, 0 means impossible to
        /// capture.
        u8 GetCaptureRate() const noexcept;

        /// @brief Checks whether the species contains a variant with a specific name.
        /// @param key The name of the variant that's being looked for.
        /// @return True if the species contains the variant, false otherwise.
        [[nodiscard]] bool HasVariant(const ArbUt::BasicStringView& key) const noexcept;
        /// @brief Checks whether the species contains a variant with a specific name.
        /// @param hash The string hash of the variant that's being looked for. This hash can be retrieved from the
        /// StringView class.
        /// @return True if the species contains the variant, false otherwise.
        [[nodiscard]] bool HasVariant(u32 hash) const noexcept;
        /// @brief Try to get a variant of the species with a specific name.
        /// @param name The name of the variant that's being looked for.
        /// @param out If a variant is found, it will be put in this variable. If not, this will remain unchanged.
        /// @return True if the species contains the variant, false otherwise.
        [[nodiscard]] std::optional<ArbUt::BorrowedPtr<const SpeciesVariant>>
        TryGetVariant(const ArbUt::BasicStringView& name) const noexcept;
        /// @brief Try to get a variant of the species with a specific name.
        /// @param name The string hash of the variant that's being looked for. This hash can be retrieved from the
        /// StringView class.
        /// @param out If a variant is found, it will be put in this variable. If not, this will remain unchanged.
        /// @return True if the species contains the variant, false otherwise.
        [[nodiscard]] std::optional<ArbUt::BorrowedPtr<const SpeciesVariant>> TryGetVariant(u32 hash) const noexcept;
        /// @brief Returns a variant with a specific name. Throws if no variant with the name exists.
        /// @param key The name of the variant that's being looked for.
        /// @return The specified variant.
        [[nodiscard]] ArbUt::BorrowedPtr<const SpeciesVariant> GetVariant(const ArbUt::BasicStringView& key) const;
        /// @brief Returns a variant with a specific name. Throws if no variant with the name exists.
        /// @param key The string hash of the variant that's being looked for.
        /// @return The specified variant.
        [[nodiscard]] ArbUt::BorrowedPtr<const SpeciesVariant> GetVariant(u32 key) const;
        /// @brief Returns a random gender based on the gender ratio of the species.
        /// @param rand A random number generator class.
        /// @return A random gender. If gender ratio is -1 this will return Genderless, otherwise it will be either male
        /// or female.
        [[nodiscard]] Gender GetRandomGender(ArbUt::Random& rand) const noexcept;

        /// @brief Appends a variant to the species.
        /// @param name The unique name of the new variant.
        /// @param variant The new variant to add.
        void SetVariant(const ArbUt::StringView& name, const SpeciesVariant* non_null variant);

        /// @brief Returns a list of variants of the species.
        /// @return A list of variants of the species.
        const ArbUt::List<ArbUt::BorrowedPtr<const SpeciesVariant>>& GetVariantsIterator() const;

        /// @brief Checks whether the species has a specific flag.
        /// @param key The flag to check for.
        /// @return True if the species has the flag, false otherwise.
        bool HasFlag(const ArbUt::StringView& key) const noexcept;
        /// @brief Checks whether the species has a specific flag.
        /// @param keyHash The flag to check for.
        /// @return True if the species has the flag, false otherwise.
        bool HasFlag(u32 keyHash) const noexcept;
    };
}

#endif // CREATURELIB_CREATURESPECIES_HPP
